;; Generated package description from dired-preview.el  -*- no-byte-compile: t -*-
(define-package "dired-preview" "0.1.1.0.20240506.70544" "Automatically preview file at point in Dired" '((emacs "27.1")) :commit "5b24955710cd24ce17d4baaafe1a7a6de5e6377d" :authors '(("Protesilaos Stavrou" . "info@protesilaos.com")) :maintainer '("Protesilaos Stavrou" . "info@protesilaos.com") :keywords '("files" "convenience") :url "https://github.com/protesilaos/dired-preview")
